/*
 *	LASI/ASP/WAX/EISA interrupt manager
 *
 *	(c) Copyright 1999 Red Hat Software
 *	Portions (c) Copyright 1999 The Puffin Group Inc.
 *	Portions (c) Copyright 1999 Hewlett-Packard
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *	by Alan Cox <alan@redhat.com> and 
 * 	   Alex deVries <adevries@thepuffingroup.com>
 *	   Helge Deller <deller@gmx.de>
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/bitops.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/gsc.h>

#include "busdevice.h"


#undef DEBUG

#ifdef DEBUG
#define DEBPRINTK printk
#else
#define DEBPRINTK(x,...)
#endif


static struct pa_iodc_driver busdev_drivers_for[] __initdata = {
/* I'm pretty sure this covers them all. */
  {HPHW_BA, 0x0, 0, 0x00081, 0, 0,
	DRIVER_CHECK_HWTYPE + DRIVER_CHECK_SVERSION,
	"Lasi", "generic", (void *) lasi_init_chip},
  {HPHW_BA, 0x0, 0, 0x00070, 0, 0,
	DRIVER_CHECK_HWTYPE + DRIVER_CHECK_SVERSION,
	"ASP", "generic", (void *) asp_init_chip},
  { 0 }
};


/* IRQ bits must be numbered from Most Significant Bit */
#define BUSDEV_FIX_IRQ(x)	(31-(x))
#define BUSDEV_MASK_IRQ(x)	(1<<(BUSDEV_FIX_IRQ(x)))


static struct busdevice *busdev_list = NULL;


int busdevice_alloc_irq( struct hp_device *dev )
{
	struct busdevice *busdev = busdev_list;
	int	irq;

	if (!busdev)
	{
		printk(KERN_ERR "%s(0x%p): No LASI/ASP/WAX found in system yet !\n",
			__FUNCTION__, dev);
		return 0;
	}

	/* See if this Device belongs to a LASI/ASP/WAX we know about */
	/* deller: Changed to test only the 3 highest nibbles of the
	    address-range, since ASP uses hpa of 0xf080yyyy, but devices are 
	    at address 0xf082yyyy ! */
	while (busdev &&
		(((unsigned long) busdev->hpa & ~0xfffff) != ((unsigned long) dev->hpa & ~0xfffff)))
	{
		busdev = busdev->next;
	}

	if (!busdev)
	{
		printk(KERN_WARNING "%s(0x%p): No known LASI/ASP/WAX owns device at 0x%p !\n",
			__FUNCTION__, dev, dev->hpa);
		return 0;
	}

	/* find the correspondig IRQ */	
	irq = busdev->find_irq(busdev, dev);
	
	if (irq < 0) {
	    printk(KERN_ERR "%s(0x%p): %s has never seen the HPA offset of 0x%lx.\n", 
			    __FUNCTION__, dev->hpa, busdev->name, 
			    ((unsigned long) dev->hpa) & 0xffff );
	    return (0);
	}

	DEBPRINTK (KERN_INFO "%s(0x%p) on %s 0x%x + %d = %d\n", 
		    __FUNCTION__, dev->hpa, busdev->name, 
		    busdev->busdev_region->data.irqbase, irq, 
		    busdev->busdev_region->data.irqbase + irq);

	return busdev->busdev_region->data.irqbase + irq;
}


/*
 *	Deliver interrupts
 */
 
void busdev_barked(int busdev_irq, void *dev, struct pt_regs *regs)
{
	u32 irq;
	struct busdevice *busdev = (struct busdevice *) dev;

	/* 
	    Don't need to protect OFFSET_IRR with spinlock since this is
	    the only place it's touched.
	    Protect busdev_region by disabling this region's interrupts,
	    modifying the region, and then re-enabling the region.
	*/

	irq = gsc_readl(busdev->hpa+OFFSET_IRR);
	if(irq==0)
		printk(KERN_ERR "%s: barking without apparent reason.\n", busdev->name);
	else {
		DEBPRINTK ("%s (0x%x) barked, mask=0x%x, irq=%d\n", 
		    busdev->name, busdev->busdev_region->data.irqbase, 
		    irq, BUSDEV_FIX_IRQ(ffs(irq))+1 );

		do_irq_mask((unsigned long) irq, busdev->busdev_region, regs);
	}
}

static void
busdev_disable_irq(void *irq_dev, int irq)
{
	/* Disable the IRQ line by clearing the bit in the IMR */
	u32 imr = gsc_readl(BUSDEV_DEV(irq_dev)->hpa+OFFSET_IMR);
	imr &= ~(BUSDEV_MASK_IRQ(irq));

	DEBPRINTK( KERN_WARNING "%s(%p, %d) %s: IMR 0x%x\n", 
		    __FUNCTION__, irq_dev, irq, BUSDEV_DEV(irq_dev)->name, imr);

	gsc_writel(imr, BUSDEV_DEV(irq_dev)->hpa+OFFSET_IMR);
}


static void
busdev_enable_irq(void *irq_dev, int irq)
{
	/* Enable the IRQ line by setting the bit in the IMR */
	char *addr = (char*) (BUSDEV_DEV(irq_dev)->hpa + OFFSET_IMR);
	u32 imr = gsc_readl(addr);
	imr |= BUSDEV_MASK_IRQ(irq);

	DEBPRINTK (KERN_WARNING "%s(%p, %d) %s: IMR 0x%x\n", 
		    __FUNCTION__, irq_dev, irq, BUSDEV_DEV(irq_dev)->name, imr);

	gsc_writel(imr, addr);
//	gsc_writel(~0L, addr);

/* FIXME: read IPR to make sure the IRQ isn't already pending.
**   If so, we need to read IRR and manually call do_irq_mask().
**   This code should be shared with busdev_unmask_irq().
*/
}

static void
busdev_mask_irq(void *irq_dev, int irq)
{
/* FIXME: Clear the IMR bit in busdev for that IRQ */
}

static void
busdev_unmask_irq(void *irq_dev, int irq)
{
/* FIXME: Read IPR. Set the IMR bit in busdev for that IRQ.
   call do_irq_mask() if IPR is non-zero
*/
}

struct irq_region_ops busdev_irq_ops = {
	busdev_disable_irq,	/* disable_irq */
	busdev_enable_irq,	/* enable_irq */
	busdev_mask_irq,	/* void (* mask_irq)(void *dev, int irq) */
	busdev_unmask_irq	/* void (* unmask_irq)(void *dev, int irq) */
};


#define DEBUG_BUSDEVICE( dev ) \
    do { \
	printk(KERN_WARNING "%s IRQ %d EIM 0x%x", dev->name, dev->parent_irq, dev->eim); \
	if (gsc_readl(dev->hpa + OFFSET_IMR)) \
	    printk(KERN_WARNING "  IMR is non-zero! (0x%x)", gsc_readl(dev->hpa + OFFSET_IMR)); \
	printk("\n"); \
    } while (0)


int register_busdevice( struct hp_device *gsc_parent, 
			 struct busdevice *busdev )
{
	struct resource *res;

	busdev->spinlock = SPIN_LOCK_UNLOCKED;
	busdev->gsc  = gsc_parent;

	/* the IRQs we simulate */
	busdev->busdev_region = alloc_irq_region(32, &busdev_irq_ops,
			    IRQ_REG_MASK|IRQ_REG_DIS, busdev->name, busdev);
	if(!busdev->busdev_region)
		return (-ENOMEM);

	/* allocate resource region */
	res = kmalloc(sizeof(struct resource), GFP_KERNEL);
	if (res) {
	    res->name = busdev->name;
	    res->start = (unsigned long) busdev->hpa;
	    res->end = res->start + 0x100000 - 1;
	    res->flags = IORESOURCE_IO; 	/* do not mark it busy ! */
	    res->child = NULL;
	    request_resource(&ioport_resource, res);
	    request_region(res->start, OFFSET_IAR + sizeof(int), "reserved");
	}
	
#if 0
	DEBUG_BUSDEVICE( busdev );
#endif

	busdev->next = busdev_list;
	busdev_list  = busdev;
	
	return 0;
}



/*
 *	Initialize all controller chips (LASI/ASP/WAX) and attached devices.
 */
extern int parport_gsc_init (void);
extern void register_lan_drivers(void);
extern void register_hil_drivers(void);
extern int lasi_psaux_init(void);
extern int gsc_ps2_init(void);



int __init busdevices_init(void)
{
	register_driver(busdev_drivers_for);

#ifdef CONFIG_GSC_WAX
	register_wax_driver();
#endif

#ifdef CONFIG_PARPORT_GSC 
	parport_gsc_init();	/* parallel port driver */
#endif	

#ifdef CONFIG_LASI_82596
	register_lan_drivers(); /* LASI/ASP LAN (i82596) */
#endif

#ifdef CONFIG_GSC_PS2
	gsc_ps2_init();		/* PS aux driver */
#endif

#ifdef CONFIG_HIL
        register_hil_drivers();	/* Human Interface Loop driver */
#endif

#ifdef CONFIG_SOUND_GSC_HARMONY
	register_harmony_drivers();
#endif

	return 0;
}
