/*
 *	The Gecko device driver
 *
 *	(c) Copyright 1999 The Puffin Group Inc.
 *	(c) Copyright 1999 SuSE GmbH (Philipp Rumpf, prumpf@tux.org)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 * This is the GSC code.
 * 
 *      by Alex deVries <adevries@thepuffingroup.com>
 * 
 */

#include <linux/errno.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <asm/gsc.h>
#include <asm/led.h>

#include <asm/io.h>
#include <asm/delay.h>
#include <linux/malloc.h>
#include <linux/smp_lock.h>
#include <linux/interrupt.h>

#include <asm/page.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/smp.h>
#include <linux/spinlock.h>

#include <asm/irq.h>
#include <asm/machdep.h>

#include <asm/processor.h>	/* for boot_cpu_data */
/*
** Intended for ccio/sba/cpu statistics under /proc/bus/{runway|gsc}
*/
struct proc_dir_entry * proc_runway_root = NULL;
struct proc_dir_entry * proc_gsc_root = NULL;


/* values available for io_command: */
#define	CMD_STOP	0	/* halt any I/O, enable diagnostic access */
#define	CMD_FLUSH	1	/* abort DMA */
#define	CMD_CHAIN	2	/* initiate DMA */
#define	CMD_CLEAR	3	/* clear errors */
#define	CMD_RESET	5	/* reset any module */
#define CMD_PORT	32	/* set port bit on bus convertor */

u8 _gsc_readb(void *addr)
{
	long flags;
	u8 ret;

	asm volatile(
	"	rsm	2,%0
		ldbx	0(%2),%1
		mtsm	%0"
	: "=&r" (flags), "=r" (ret) : "r" (addr) );

	return ret;
}

u16 _gsc_readw(void *addr)
{
	long flags;
	u16 ret;

	asm volatile(
	"	rsm	2,%0
		ldhx	0(%2),%1
		mtsm	%0"
	: "=&r" (flags), "=r" (ret) : "r" (addr) );

	return ret;
}

u32 _gsc_readl(void *addr)
{
	long flags;
	u32 ret;

	asm volatile(
	"	rsm	2,%0
		ldwx	0(%2),%1
		mtsm	%0"
	: "=&r" (flags), "=r" (ret) : "r" (addr) );

	return ret;
}

u64 _gsc_readq(void *addr)
{
	u64 ret;
#if BITS_PER_LONG > 32
	long flags;
	asm volatile(
	"	rsm	2,%0
		ldd	0(%2),%1
		mtsm	%0"
	: "=&r" (flags), "=r" (ret) : "r" (addr) );
#else
	/* two reads may have side effects.. */
	ret = _gsc_readl(addr);
	ret <<= 32;
	ret |= _gsc_readl(addr+4);
#endif
	return ret;
}

void _gsc_writeb(u8 val, void *addr)
{
	long flags;
	asm volatile(
	"	rsm	2,%0
		stbs	%1,0(%2)
		mtsm	%0"
	: "=&r" (flags) :  "r" (val), "r" (addr) );
}
	
void _gsc_writew(u16 val, void *addr)
{
	long flags;
	asm volatile(
	"	rsm	2,%0
		sths	%1,0(%2)
		mtsm	%0"
	: "=&r" (flags) :  "r" (val), "r" (addr) );
}

void _gsc_writel(u32 val, void *addr)
{
	long flags;
	asm volatile(
	"	rsm	2,%0
		stws	%1,0(%2)
		mtsm	%0"
	: "=&r" (flags) :  "r" (val), "r" (addr) );
}

void _gsc_writeq(u64 val, void *addr)
{
#if BITS_PER_LONG > 32
	long flags;
	asm volatile(
	"	rsm	2,%0
		std	%1,0(%2)
		mtsm	%0"
	: "=&r" (flags) :  "r" (val), "r" (addr) );
#else
	/* two writes may have side effects.. */
	_gsc_writel(val>>32, addr);
	_gsc_writel(val,     addr+4);
#endif
}


int gsc_alloc_irq(struct gsc_irq *i)
{
	int irq = txn_alloc_irq();
	if(irq < 0) {
		printk("cannot get irq\n");
		return irq;
	}

	i->txn_addr = txn_alloc_addr(irq);
	i->txn_data = txn_alloc_data(irq, GSC_EIM_WIDTH);
	i->irq = irq;

	return irq;
}


int gsc_claim_irq(struct gsc_irq *i, int irq)
{
	int c = irq;

	irq += IRQ_FROM_REGION(CPU_IRQ_REGION); /* virtualize the IRQ first */

	irq = txn_claim_irq(irq);
	if(irq < 0) {
		printk("cannot claim irq %d\n", c);
		return irq;
	}

	i->txn_addr = txn_alloc_addr(irq);
	i->txn_data = txn_alloc_data(irq, GSC_EIM_WIDTH);
	i->irq = irq;

	return irq;
}


void __init gsc_init(void)
{
#if defined(CONFIG_PCI)
#if defined(CONFIG_GSC_DINO)
	extern void dino_init(void);
#endif
	/* FIXME: This isn't GSC related.
	** In fact, these boxes don't have GSC at all.
	** It's just a convient place to put these
	** calls for testing pre-release code.
	*/

#if defined(CONFIG_IOSAPIC)
	extern void iosapic_init(void);
#endif
#if defined(CONFIG_PCI_LBA)
	extern void lba_init(void);
#endif
#if defined(CONFIG_IOMMU_SBA)
	extern void sba_init(void);
#endif
#endif	/* CONFIG_PCI */
#if defined(CONFIG_IOMMU_CCIO)
	extern void ccio_init(void);
#endif
#ifdef CONFIG_GSC_LASI
	extern void busdevices_init(void);
#endif

	/*
	** Can't call proc_mkdir() until after proc_root_init() has been
	** called by start_kernel(). In other words, this code can't
	** live in arch/.../setup.c because start_parisc() calls
	** start_kernel().
	*/
	switch (boot_cpu_data.cpu_type) {
	case pcxl:
	case pcxl2:
		if (NULL == proc_gsc_root)
		{
			proc_gsc_root = proc_mkdir("bus/gsc", 0);
		}
		break;
        case pcxt_:
        case pcxu:
        case pcxu_:
        case pcxw:
        case pcxw_:
                if (NULL == proc_runway_root)
                {
                        proc_runway_root = proc_mkdir("bus/runway", 0);
                }
                break;
	}


#ifdef CONFIG_GSC_LASI
	busdevices_init();
#endif

#if defined(CONFIG_PCI)
#if defined(CONFIG_GSC_DINO)
	dino_init();
#endif
#if defined(CONFIG_IOSAPIC)
	iosapic_init();
#endif
#if defined(CONFIG_PCI_LBA)
	lba_init();
#endif
#if defined(CONFIG_IOMMU_SBA)
	sba_init();
#endif
#endif

#if defined(CONFIG_IOMMU_CCIO)
	ccio_init();
#endif

#ifdef CONFIG_CHASSIS_LCD_LED
	register_led_regions();	/* register LED port info in procfs */
#endif
}
